;********************************************************************************************************
;                                              uC/OS-II
;                                        The Real-Time Kernel
;
;                    Copyright 1992-2021 Silicon Laboratories Inc. www.silabs.com
;
;                                 SPDX-License-Identifier: APACHE-2.0
;
;               This software is subject to an open source license and is distributed by
;                Silicon Laboratories Inc. pursuant to the terms of the Apache License,
;                    Version 2.0 available at www.apache.org/licenses/LICENSE-2.0.
;
;********************************************************************************************************

;********************************************************************************************************
;
;                                       MCS51-STC8 Specific code
;                                           keil C51 Compiler
;
; Filename : os_cpu_a.A51
; Version  : V2.93.01
;修改版本： V1.03
;修改时间： 2023-08-27
;作者：熊仔
;QQ：   616421684
;设置DPTR0自动递增模式,加快拷贝数据速度
;注意：统一用寄存器组0，加快出入栈速度
;********************************************************************************************************

$NOMOD51
    
    
;使用DPTR自动递增功能优化代码.大概节省1us时间
;注意这部分代码不能进行硬件仿真单步调试   
;因为STC8H的串口仿真和USB仿真都是monitor仿真，是需要运行仿真监控程序的
;仿真监控程序里面有使用DPTR相关的指令，会影响用户DPTR的自动递增功能
DPTR_AUTO_INC       EQU     0

;IE  DATA    0A8H
EA  BIT     0A8H.7
SP  DATA    081H
B   DATA    0F0H
ACC DATA    0E0H
DPH DATA    083H
DPL DATA    082H
PSW DATA    0D0H
TH0 DATA    08CH
TL0 DATA    08AH
P2  DATA    0A0H
DPS DATA    0E3H
TA  DATA    0AEH


    
        NAME OS_CPU_A	;模块名
        
;定义重定位段
;1：无参函数 ?PR?函数名?文件名
;2：有参函数 ?PR?_函数名?文件名
;3：再入函数 ?PR?_?函数名?文件名
?PR?OSStartHighRdy?OS_CPU_A    SEGMENT CODE
?PR?OSCtxSw?OS_CPU_A           SEGMENT CODE
?PR?OSIntCtxSw?OS_CPU_A        SEGMENT CODE
?PR?OSTickISR?OS_CPU_A         SEGMENT CODE

        
;声明引用全局变量和外部子程序
        EXTRN DATA  (?C_XBP)	;仿真堆栈指针用于可重入局部变量保存

        EXTRN DATA (OSTCBCur)
        EXTRN DATA (OSTCBHighRdy)
        EXTRN BIT (OSRunning)
        EXTRN DATA (OSPrioCur)
        EXTRN DATA (OSPrioHighRdy)      
        EXTRN BIT (_bEA)           ;切换任务前备份EA的值，最后退出的时候，恢复EA

        EXTRN CODE  (_?OSTaskSwHook)
        EXTRN CODE  (_?OSIntEnter)
        EXTRN CODE  (_?OSIntExit)
        EXTRN CODE  (_?OSTimeTick)

            
;对外声明3个不可重入函数
        PUBLIC OSStartHighRdy
        PUBLIC OSCtxSw
        PUBLIC OSIntCtxSw

    
;分配堆栈空间，这里直接使用C51分配的堆栈空间，栈底指针?STACK由C51自动生成，可以查看启动文件START.A51，默认分配大小是1，
;可以修改分配的空间大小比如40H，64字节，避免用户使用idata过多空间不足引起堆栈溢出。
;链接的时候会和START.A51声明的?STACK合并
?STACK SEGMENT IDATA
        RSEG ?STACK

;定义压栈出栈宏
PUSHALL    MACRO
        PUSH PSW
        PUSH ACC
        PUSH B
        PUSH DPL
        PUSH DPH
        PUSH AR0
        PUSH AR1
        PUSH AR2
        PUSH AR3
        PUSH AR4
        PUSH AR5
        PUSH AR6
        PUSH AR7
        ENDM
    
POPALL    MACRO
        POP  AR7
        POP  AR6
        POP  AR5
        POP  AR4
        POP  AR3
        POP  AR2
        POP  AR1
        POP  AR0
        POP  DPH
        POP  DPL
        POP  B
        POP  ACC
        POP  PSW
        ENDM
    
;--------------------
;子程序
;--------------------

;   --------------------------------------------------
;   ---------------- OSStartHighRdy() ----------------
;   --------------------------------------------------

        RSEG ?PR?OSStartHighRdy?OS_CPU_A

;调用用户扩展函数OSTaskSwHook()
;OSRunning=TRUE
;取得将要恢复的任务的堆栈指针：Stack pointer=OSTCBHighRdy->OSTCBStkPtr
;将所有寄存器内容从任务栈中弹出来
;执行返回指令RET

OSStartHighRdy:

        USING 0 ;
        CLR EA    ;关中断，避免用户初始化不慎打开了EA，返回任务的时候，开中断。
        SETB _bEA  ;_bEA置1
        LCALL _?OSTaskSwHook
        ;OSRunning=TRUE
        SETB OSRunning          ;使用bit加速
IF DPTR_AUTO_INC <> 0
        ;设置DPTR0自动递增模式
        MOV     TA,#0AAH        ;写入DPS写保护触发命令1
        MOV     TA,#55H         ;写入DPS写保护触发命令2
        MOV     DPS,#00001000B  ;DPTR0递增,单独使能AU0,并选择DPTR0
ENDIF
;--------------------
;OSCtxSw_in
;--------------------

OSCtxSw_in:
        ;这个时候 OSTCBHighRdy = OSTCBCur，所以这里操作OSTCBHighRdy或者OSTCBCur都是正确的
        ;OSTCBHighRdy ==> DPTR 获得当前就绪的最高优先级任务的TCB指针
        ;由于定义OSTCBHighRdy是指向XDATA区的，所以占2字节。+0高8位数据+1低8位数据 
        MOV   DPH, OSTCBHighRdy
        MOV   DPL, OSTCBHighRdy+1
        ;OSTCBHighRdy->OSTCBStkPtr ==> DPTR 获得用户堆栈指针
        ;typedef unsigned char xdata    OS_STK;堆栈指针声明xdata，占用2字节，屏蔽INC  DPTR
;        INC  DPTR       ;指针占3B。+0类型 +1高8位数据 +2低8位数据
        MOVX A,@DPTR     ;OSTCBStkPtr是void指针
        MOV  R0,A
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOVX A,@DPTR
        MOV  DPL,A   ;优化
        MOV  DPH,R0
        ;MOV  R1,A
        ;MOV  DPH,R0
        ;MOV  DPL,R1

        ;*UserStkPtr ==> R5 用户堆栈起始地址内容(即用户堆栈长度放在此处)   
        MOVX A,@DPTR    ;用户堆栈中是unsigned char类型数据
        MOV  R5,A       ;R5=用户堆栈长度

        ;恢复现场堆栈内容
        MOV  R0,#?STACK-1

;--------------------
;restore_stack
;--------------------
        
restore_stack:
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        INC  R0
        MOVX A,@DPTR
        MOV  @R0,A
        DJNZ R5,restore_stack
    
        ;恢复堆栈指针SP
        MOV  SP,R0
    
        ;恢复仿真堆栈指针?C_XBP        
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOVX A,@DPTR
        MOV  ?C_XBP,A     ;?C_XBP 仿真堆栈指针高8位
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOVX A,@DPTR
        MOV  ?C_XBP+1,A   ;?C_XBP 仿真堆栈指针低8位
IF DPTR_AUTO_INC <> 0
        ;设置DPTR0默认模式
        MOV     TA,#0AAH  ;写入DPS写保护触发命令1
        MOV     TA,#55H   ;写入DPS写保护触发命令2
        MOV     DPS,#0H   ;DPS恢复默认值
ENDIF
        POP  AR7
        POP  AR6
        POP  AR5
        POP  AR4
        POP  AR3
        POP  AR2
        POP  AR1
        POP  AR0
        POP  DPH
        POP  DPL
        POP  B
        POP  ACC
        
        MOV  C,_bEA
        MOV  EA,C
        
        POP  PSW              ;应该放最后，前面2条指令用到CY
;;        CLR P2.6            ;测试信号***************************************************
;;        CLR P2.1            ;测试信号***************************************************
;;        CLR P2.3            ;测试信号***************************************************
        RET
        
        
;   --------------------------------------------------
;   ------------------- OSCtxSw() --------------------
;   --------------------------------------------------

        RSEG ?PR?OSCtxSw?OS_CPU_A

;保存处理器寄存器
;将当前任务的堆栈指针保存到当前任务的OS_TCB中：OSTCBCur->OSTCBStkPtr=Stack pointer
;调用用户扩展函数OSTaskSwHook()
;OSTCBCur=OSTCBHighRdy
;OSPrioCur=OSPrioHighRdy
;得到需要恢复的任务的堆栈指针：Stack pointer=OSTCBHighRdy->OSTCBStkPtr
;将所有处理器寄存器从新任务的堆栈中恢复出来
;执行返回指令

OSCtxSw:  

        USING 0  
;SETB    P2.3           ;测试信号***************************************************
        
        MOV  C,EA   ;先保存中断状态位
        CLR  EA     ;关中断
        MOV  _bEA,C ;中断状态位保存到备份变量_bEA
        PUSHALL
        
;--------------------
;OSIntCtxSw_in
;--------------------
        
OSIntCtxSw_in:
IF DPTR_AUTO_INC <> 0
        ;设置DPTR0自动递增模式
        MOV     TA,#0AAH        ;写入DPS写保护触发命令1
        MOV     TA,#55H         ;写入DPS写保护触发命令2
        MOV     DPS,#00001000B  ;DPTR0递增,单独使能AU0,并选择DPTR0
ENDIF
        ;获得堆栈长度和起始地址，任务栈结构示意图(OS_CPU_C.C)
        MOV  A,SP
        CLR  C  ;进位标志位Cy清零
        ;        SUBB A,#OSStkStart	;A <- A-OSStkStart-Cy
        SUBB A,#?STACK-1
        MOV  R5,A     ;获得堆栈长度        

        ;OSTCBCur ==> DPTR  获得当前TCB指针
        MOV  DPH, OSTCBCur
        MOV  DPL, OSTCBCur+1    

        ;OSTCBCur->OSTCBStkPtr ==> DPTR  获得用户堆栈指针
        ;typedef unsigned char xdata	OS_STK;堆栈指针声明xdata，占用2字节，屏蔽INC  DPTR
;        INC  DPTR        ;指针占3B。+0类型 +1高8位数据 +2低8位数据
        MOVX A,@DPTR      ;OSTCBStkPtr是void指针
        MOV  R0,A
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOVX A,@DPTR
        MOV  DPL,A  ;优化
        MOV  DPH,R0
        ;MOV  R1,A
        ;MOV  DPH,R0
        ;MOV  DPL,R1

        ;保存堆栈长度
        MOV  A,R5
        MOVX @DPTR,A

        MOV  R0,#?STACK-1     ;获得堆栈起始地址

;--------------------
;save_stack
;--------------------

save_stack:
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        INC  R0
        MOV  A,@R0
        MOVX @DPTR,A
        DJNZ R5,save_stack
        
        ;保存仿真堆栈指针?C_XBP
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOV  A,?C_XBP   ;?C_XBP 仿真堆栈指针高8位
        MOVX @DPTR,A
IF DPTR_AUTO_INC <> 0
        ;DPTR0 自动递增 不需要INC  DPTR
ELSE
        INC  DPTR       ;
ENDIF
        MOV  A,?C_XBP+1	;?C_XBP 仿真堆栈指针低8位
        MOVX @DPTR,A        
    
        ;调用用户程序
        LCALL _?OSTaskSwHook

        ;OSTCBCur = OSTCBHighRdy
        MOV OSTCBCur,OSTCBHighRdy 
        MOV OSTCBCur+1,OSTCBHighRdy+1 
        ;OSPrioCur = OSPrioHighRdy 使用这两个变量主要目的是为了使指针比较变为字节比较，以便节省时间
        MOV OSPrioCur,OSPrioHighRdy   ;改用data变量，提速
        LJMP OSCtxSw_in

;   -------------------------------------------
;   --------------- OSIntCtxSw() --------------
;   -------------------------------------------

        RSEG ?PR?OSIntCtxSw?OS_CPU_A

;调整堆栈指针来去掉在调用过程中压入堆栈的多余内容
;巧妙RETI返回到下一条指令
;将当前任务堆栈指针保存到当前任务的OS_TCB中：OSTCBCur->OSTCBStkPtr=Stack pointer
;调用用户扩展函数OSTaskSwHook()
;OSTCBCur=OSTCBHighRdy
;OSPrioCur=OSPrioHighRdy
;得到需要恢复的任务的堆栈指针：Stack pointer=OSTCBHighRdy->OSTCBStkPtr
;将所有处理器寄存器从新任务的堆栈中恢复出来
;执行返回指令

OSIntCtxSw:

        USING 0
;        SETB P2.1        ;测试信号***************************************************
;        CLR  P2.0        ;测试信号***************************************************
         CLR  EA    ;关中断        ;由于调用OSIntCtxSw前，退出了临界区，需要关中断
         SETB _bEA        ;中断切换的，直接置1就行
         
        ;调整SP指针去掉在调用OSIntExit()、OSIntCtxSw()过程中压入堆栈的多余内容
        ;SP=SP-4
        MOV  A,SP
        CLR  C
        SUBB A,#4
        MOV  SP,A
 
        ;提前退出中断，统一用RET返回
        ;方法1：直接把  OSIntCtxSw_in函数地址入栈，由RETI中断返回，跳转到OSIntCtxSw_in函数,共6T
        MOV DPTR,#OSIntCtxSw_in
        PUSH DPL
        PUSH DPH
        RETI
         ;方法2：;巧妙的利用函数返回RETI执行了中断完毕，跳转到下一条语句。这样任务可以无忧使用RET作为返回，共9T。
;        LCALL    RE_RETI
;        LJMP OSIntCtxSw_in
;RE_RETI:	RETI


EXTRN CODE  (?C?ADDXBP)
?PR?OS_CRITICAL_METHOD_3_FIX?OS_CPU_A    SEGMENT CODE
PUBLIC OS_CRITICAL_METHOD_3_FIX
RSEG ?PR?OS_CRITICAL_METHOD_3_FIX?OS_CPU_A
    
OS_CRITICAL_METHOD_3_FIX:
;调整OS_ENTER_CRITICAL()过程中cpu_sr压入仿真栈的内容，?C_XBP++操作
        MOV      DPTR,#0x01
        LCALL     ?C?ADDXBP
        RET
    
    
;编译有警告，还是用C函数实现   
;?PR?OS_CPU_SR_Save?OS_CPU_A    SEGMENT CODE
;?PR?_OS_CPU_SR_Restore?OS_CPU_A SEGMENT CODE
;PUBLIC OS_CPU_SR_Save
;PUBLIC _OS_CPU_SR_Restore

;RSEG ?PR?OS_CPU_SR_Save?OS_CPU_A
;OS_CPU_SR_Save:
;                MOV      R7,IE                    ; Get current state of interrupts disable flag
;                CLR EA                            ; Disable interrupts
;                RET                               ; Return original SREG value in R16

;RSEG ?PR?_OS_CPU_SR_Restore?OS_CPU_A
;_OS_CPU_SR_Restore:
;                MOV     IE,R7                    ; Restore SREG
;                RET                              ; Return
                
                
;   -------------------------------------------
;   --------------- OSTickISR() ---------------
;   -------------------------------------------

;CSEG [AT 绝对地址表达式] //绝对代码段
;DSEG [AT 绝对地址表达式] //内部绝对数据段
;XSEG [AT 绝对地址表达式] //外部绝对数据段
;ISEG [AT 绝对地址表达式] //内部间接寻址绝对数据段
;BSEG [AT 绝对地址表达式] //绝对位寻址段
;        RSEG ?PR?_?OSTickISR?OS_CPU_A

;EA = 0 关中断
;保存处理器寄存器的值
;调用OSIntEnter()或是将OSIntNesting加1
;EA = 1 允许中断嵌套
;调用OSTimeTick()
;调用OSIntExit()
;恢复处理器寄存器的值
;执行中断返回指令



; ==== 定时器0 中断服务程序 =============
OSTickISR:                
        USING 0     ;工作寄存器0
        CLR EA  ;调用OSIntEnter之前先关中断
;SETB P2.6; 测量OSTickISR执行时间     ;测试信号***************************************************
        PUSHALL ;现场保护
        LCALL _?OSIntEnter	;通知内核进入中断     
        SETB EA    ;允许中断嵌套
;SETB P2.7;  测量OSTimeTick          ;测试信号***************************************************
       	LCALL _?OSTimeTick  ;调用中断服务子程序
;CLR  P2.7                           ;测试信号***************************************************

;SETB P2.0;  测量OSIntExit执行时间    = ;测试信号***************************************************
        LCALL _?OSIntExit   ;通知内核退出中断
;CLR  P2.0                           ;测试信号***************************************************
        POPALL  ;恢复现场 
;CLR  P2.6                           ;测试信号***************************************************
        RETI



; **** 中断矢量区域 *************
; ==== 定时器0 中断服务程序 =============
        CSEG AT 000BH	;定时器T0中断
        LJMP OSTickISR

;--------------------
END
;--------------------
