/*
*********************************************************************************************************
*                                              uC/OS-II
*                                        The Real-Time Kernel
*
*                    Copyright 1992-2021 Silicon Laboratories Inc. www.silabs.com
*
*                                 SPDX-License-Identifier: APACHE-2.0
*
*               This software is subject to an open source license and is distributed by
*                Silicon Laboratories Inc. pursuant to the terms of the Apache License,
*                    Version 2.0 available at www.apache.org/licenses/LICENSE-2.0.
*
*********************************************************************************************************
*/


/*
*********************************************************************************************************
*
*                                    MCS51-STC8 Specific code
*
* Filename : os_cpu_c.c
* Version  : V2.93.01
*修改版本： V1.03
*修改时间： 2023-08-27
*作者：熊仔
*QQ：   616421684
*********************************************************************************************************
*/

//#define   OS_CPU_GLOBALS
#include  <ucos_ii.h>
//在大多数的操作系统中，程序不允许访问地址为0的内存，因为该内存是操作系统保留的。
//然而，内存地址0有特别重要的意义，它表明该指针不指向一个可访问的内存地址，既指向地址0的指针就是空指针。
//xdata 地址0声明一个null变量占用了该地址，保证其他变量不会使用0空间，导致指针赋值为0。
//解决任务堆栈空间小的时候 OSTCBTbl数据表从0开始分配，导致  OSTCBFreeList = &OSTCBTbl[0]; 分配到一个空指针。
char xdata null _at_  0;
/*
*********************************************************************************************************
*                                        LOCAL GLOBAL VARIABLES
*********************************************************************************************************
*/

#if      (OS_VERSION >= 281) && (OS_TMR_EN > 0)
  static  INT16U  OSTmrCtr;
#endif                                                                  /* #if (OS_VERSION >= 281) && (OS_TMR_EN > 0)               */

/*
*********************************************************************************************************
*                                       OS INITIALIZATION HOOK
*                                            (BEGINNING)
*
* Description: This function is called by OSInit() at the beginning of OSInit().
*
* Arguments  : none
*
* Note(s)    : 1) Interrupts should be disabled during this call.
*********************************************************************************************************
*/
#if OS_CPU_HOOKS_EN > 0 && OS_VERSION > 203
void  OSInitHookBegin (void) large reentrant
{
#if OS_VERSION >= 281 && OS_TMR_EN > 0
    OSTmrCtr = 0;
#endif
}
#endif

/*
*********************************************************************************************************
*                                       OS INITIALIZATION HOOK
*                                               (END)
*
* Description: This function is called by OSInit() at the end of OSInit().
*
* Arguments  : none
*
* Note(s)    : 1) Interrupts should be disabled during this call.
*********************************************************************************************************
*/
#if OS_CPU_HOOKS_EN > 0 && OS_VERSION > 203
void  OSInitHookEnd (void) large reentrant
{
}
#endif


/*
*********************************************************************************************************
*                                          TASK CREATION HOOK
*
* Description: This function is called when a task is created.
*
* Arguments  : ptcb   is a pointer to the task control block of the task being created.
*
* Note(s)    : 1) Interrupts are disabled during this call.
*********************************************************************************************************
*/
#if OS_CPU_HOOKS_EN > 0
void  OSTaskCreateHook (OS_TCB *ptcb) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TaskCreateHook(ptcb);
#else
    ptcb = ptcb;                                                         /* Prevent compiler warning                                 */
#endif
}
#endif


/*
*********************************************************************************************************
*                                           TASK DELETION HOOK
*
* Description: This function is called when a task is deleted.
*
* Arguments  : ptcb   is a pointer to the task control block of the task being deleted.
*
* Note(s)    : 1) Interrupts are disabled during this call.
*********************************************************************************************************
*/
#if OS_CPU_HOOKS_EN > 0
void  OSTaskDelHook (OS_TCB *ptcb) large reentrant
{  EA = 1;
#if OS_APP_HOOKS_EN > 0
    App_TaskDelHook(ptcb);
#else
    ptcb = ptcb;                                                         /* Prevent compiler warning                                 */
#endif
}
#endif

/*
*********************************************************************************************************
*                                             IDLE TASK HOOK
*
* Description: This function is called by the idle task.  This hook has been added to allow you to do
*              such things as STOP the CPU to conserve power.
*
* Arguments  : none
*
* Note(s)    : 1) Interrupts are enabled during this call.
*********************************************************************************************************
*/

#if OS_CPU_HOOKS_EN > 0 && OS_VERSION >= 251
void  OSTaskIdleHook (void) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TaskIdleHook();
#endif
}
#endif

/*
*********************************************************************************************************
*                                            TASK RETURN HOOK
*
* Description: This function is called if a task accidentally returns.  In other words, a task should
*              either be an infinite loop or delete itself when done.
*
* Arguments  : ptcb      is a pointer to the task control block of the task that is returning.
*
* Note(s)    : none
*********************************************************************************************************
*/

#if OS_CPU_HOOKS_EN > 0u
void  OSTaskReturnHook (OS_TCB  *ptcb) large reentrant
{
#if OS_APP_HOOKS_EN > 0u
    App_TaskReturnHook(ptcb);
#else
    ptcb = ptcb;
#endif
}
#endif

/*
*********************************************************************************************************
*                                           STATISTIC TASK HOOK
*
* Description: This function is called every second by uC/OS-II's statistics task.  This allows your
*              application to add functionality to the statistics task.
*
* Arguments  : none
*********************************************************************************************************
*/

#if OS_CPU_HOOKS_EN > 0
void  OSTaskStatHook (void) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TaskStatHook();
#endif
}
#endif


/*
**********************************************************************************************************
*                                       INITIALIZE A TASK'S STACK
*
* Description: This function is called by either OSTaskCreate() or OSTaskCreateExt() to initialize the
*              stack frame of the task being created. This function is highly processor specific.
*
* Arguments  : task          is a pointer to the task code
*
*              p_arg         is a pointer to a user supplied data area that will be passed to the task
*                            when the task first executes.
*
*              ptos          is a pointer to the top of stack. It is assumed that 'ptos' points to the
*                            highest valid address on the stack.
*
*              opt           specifies options that can be used to alter the behavior of OSTaskStkInit().
*                            (see uCOS_II.H for OS_TASK_OPT_???).
*
* Returns    : Always returns the location of the new top-of-stack' once the processor registers have
*              been placed on the stack in the proper order.
*
* 注意:      opt           传入任务堆栈大小 实现不等长任务堆栈（熊仔）
*
* 返回值     : 函数总是返回用户堆栈空间的最低有效地址。这样修改提高了TCB换入换出的效率。
*
* 注意       : 任务堆栈结构:（参考陈是知移植版本）
* 注意      ：修改系统硬件堆栈开始地址，直接用C51自动生成的?STACK，节省一个字节。（熊仔）
*
*                                    ---------- -
*                 用户栈最高地址---->|        | |
*                                    ---------- |
*                                    |   ...  | 仿真堆栈空间
*----------                          ---------- | 每任务一个
*|OSTCBCur|               ?C_XBP---->|        | | KEIL自动处理
*----------                          ---------- -
*    |                               |空闲间隔|
*    |     -----------------------   ----------                           ----------
*    \---->|OSTCBCur->OSTCBStkPtr|   |?C_XBP低|                    SP---->|        |
*          -----------------------   ----------                           ----------
*                     |              |?C_XBP高|                           |        |
*                     |              ---------- -                         ----------
*                     |              |        | |                         |   .    |
*                     |              ---------- |                         |   .    |
*                     |              |        | |                         |   .    |
*                     |              ---------- |                         ----------
*                     |              |   .    |长度                       |        | +1
*                     |              |   .    | |                         ----------
*                     |              |   .    | |              ?STACK---->|        | 0
*                     |              ---------- |                         ----------
*                     |              |        | |          OSStkStart---->| 不关心 | -1  低地址
*                     |              ---------- -                         ----------
*                     \------------->|  长度  | 低地址                   系统硬件堆栈
*                                    ----------
*                                     用户堆栈                        长度=SP-?STACK-1
**********************************************************************************************************
*/

OS_STK  *OSTaskStkInit (void (*task)(void *pd), void *p_arg, OS_STK *ptos, INT16U opt) large reentrant
{
    OS_STK *stk;
    p_arg = p_arg;
    //opt传入任务堆栈大小 实现不等长任务堆栈
    stk	  = (OS_STK *)ptos;                 // 任务堆栈最低有效地址
    *stk++ = 15;                            // 任务堆栈长度
    *stk++ = (INT16U)task & 0xFF;           // 任务代码地址低8位
    *stk++ = (INT16U)task >> 8;             // 任务代码地址高8位
    *stk++ = 0x00;                          // PSW
    *stk++ = 0x0A;                          // ACC
    *stk++ = 0x0B;                          // B
    *stk++ = 0x00;                          // DPL
    *stk++ = 0x00;                          // DPH
    *stk++ = 0x00;                          // R0
    //R3、R2、R1用于传递任务参数p_arg，其中R3代表存储器类型，R2为高字节偏移，R1为低字节位移。
    *stk++ = (INT16U)p_arg & 0xFF;          // R1
    *stk++ = (INT16U)p_arg >> 8;            // R2
    *stk++ = (INT32U)p_arg >> 16;           // R3   存储器类型有code(0xFF),xdata(0x01),data(0x00),idata(0x00),pdata(0xFE)
    *stk++ = 0x04;                          // R4
    *stk++ = 0x05;                          // R5
    *stk++ = 0x06;                          // R6
    *stk++ = 0x07;                          // R7
                                            // 不用保存SP，任务切换时根据用户堆栈长度计算得出
    *stk++ = (INT16U)(ptos + opt) >> 8;     //?C_XBP 仿真堆栈指针高8位 opt传入任务堆栈大小
    *stk++ = (INT16U)(ptos + opt) & 0xFF;   //?C_XBP 仿真堆栈低8位 opt传入任务堆栈大小
    return (ptos);                          //返回最低地址，这里不用弹出栈顶指针是为了提高计算效率 
    //OS_TCBInit初始化堆栈 ptcb->OSTCBStkPtr= ptos;  所以第一个就是任务堆栈长度 
}

/*
*********************************************************************************************************
*                                           TASK SWITCH HOOK
*
* Description: This function is called when a task switch is performed.  This allows you to perform other
*              operations during a context switch.
*
* Arguments  : none
*
* Note(s)    : 1) Interrupts are disabled during this call.
*              2) It is assumed that the global pointer 'OSTCBHighRdy' points to the TCB of the task that
*                 will be 'switched in' (i.e. the highest priority task) and, 'OSTCBCur' points to the
*                 task being switched out (i.e. the preempted task).
*********************************************************************************************************
*/
#if (OS_CPU_HOOKS_EN > 0) && (OS_TASK_SW_HOOK_EN > 0)
void  OSTaskSwHook (void) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TaskSwHook();
#endif
}
#endif

/*
*********************************************************************************************************
*                                           OS_TCBInit() HOOK
*
* Description: This function is called by OS_TCBInit() after setting up most of the TCB.
*
* Arguments  : ptcb    is a pointer to the TCB of the task being created.
*
* Note(s)    : 1) Interrupts may or may not be ENABLED during this call.
*********************************************************************************************************
*/
#if OS_CPU_HOOKS_EN > 0 && OS_VERSION > 203
void  OSTCBInitHook (OS_TCB *ptcb) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TCBInitHook(ptcb);
#else
    ptcb = ptcb;                                                         /* Prevent compiler warning                                 */
#endif
}
#endif


/*
*********************************************************************************************************
*                                               TICK HOOK
*
* Description: This function is called every tick.
*
* Arguments  : none
*
* Note(s)    : 1) Interrupts may or may not be ENABLED during this call.
*********************************************************************************************************
*/
#if (OS_CPU_HOOKS_EN > 0) && (OS_TIME_TICK_HOOK_EN > 0)
void  OSTimeTickHook (void) large reentrant
{
#if OS_APP_HOOKS_EN > 0
    App_TimeTickHook();
#endif

#if OS_VERSION >= 281 && OS_TMR_EN > 0
    OSTmrCtr++;
    if (OSTmrCtr >= (OS_TICKS_PER_SEC / OS_TMR_CFG_TICKS_PER_SEC)) {
        OSTmrCtr = 0;
        OSTmrSignal();
    }
#endif
}
#endif
                                     
// 初始化定时器0，用于产生时钟节拍，定时器时钟1T模式，定时更准确 
void Timer0_Init(void)      //1毫秒
{
    AUXR |= 0x80;           //定时器0时钟1T模式
    TMOD &= 0xF0;           //设置定时器模式0
    //    TMOD |= 0x03;     //模式3 不可屏蔽 中断嵌套要出问题
    TL0 = -(MAIN_Fosc / OS_TICKS_PER_SEC);
    TH0 = (-(MAIN_Fosc / OS_TICKS_PER_SEC)) >> 8;
    TF0 = 0;            //清除TF0标志
    TR0 = 1;            //定时器0开始计时
    ET0 = 1;            //这里可以开启滴答定时器中断，因为执行OSStartHighRdy函数，(OSRunning = OS_TRUE)才会进行任务调度
}

bit _bEA;    //  切换任务前备份EA的值，最后退出的时候，恢复EA

#if OS_CRITICAL_METHOD == 3
OS_CPU_SR  OS_CPU_SR_Save(void)
{
    char _ie = IE;
    EA = 0;
    return _ie;
}
void OS_CPU_SR_Restore(OS_CPU_SR cpu_sr)
{
    IE = cpu_sr;
}
#endif 

#if OS_CRITICAL_METHOD==4
//加锁/解锁方式，全局变量CriticalNesting之中
INT8U      data       CriticalNesting = 0;   //养成好习惯初始化为0，虽然data区默认0.

void OSEnterCritical(void)
{
    EA = 0;
    CriticalNesting++;
}

void OSEXitCritical(void)
{
    CriticalNesting--;
    if (CriticalNesting == 0){
        EA = 1;
    }
}
#endif
